<copyright> Random classes based on double numbers.
    Written by <a href="mailto:david.wallin@telia.com">David Wallin</a>

    Copyright &copy; 2000 Pieter J. Schoenmakers.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: RandomDouble.t,v 1.2 2000/08/27 20:32:20 bruce Exp $</id>
    </copyright>

<doc> This is a Random number generator called Ranecu.  A lot of the
    actual code was borrowed from the RngPack 1.0 Java package by
    Paul Houle.

    This is a quote from the "Ranecu.java":

    "Ranecu is an advanced multiplicative linear congruential random number
    generator with a period of aproximately 10e18"

    References:

    http://www.msc.cornell.edu/~houle/rngpack

    F. James, "Comp. Phys. Comm." 60 1990  p 329-344
    P. L`Ecuyer, "Commun. ACM." 1988 1988 p 742.  </doc>
implementation class
Ranecu: State, DoubleRandom
{
  <doc> Default seeds.  </doc>
  const DEFSEED1 = 12345;
  const DEFSEED2 = 67890;
}

<doc> Return a newly allocated instance, initialized with a random seed.
    </doc>
instance (id) (r)
  new
{
  r = [self alloc];

  int s = int (2.0e8 * [Date relativeTimeIntervalSinceNow]) ^ [r hashq];
  = [r initWithSeed s];
}

<doc> Return a newly allocated instance, initialized with a default seed.
    </doc>
instance (id)
  newWithDefault
{
  = [[self alloc] init];
}

<doc> Return a newly allocated instance, initialized with the specified
    {seed}.  </doc>
instance (id)
  newWithSeed int seed
{
  = [[self alloc] initWithSeed seed];
}

end;

implementation instance
Ranecu
{
  <doc> The seed.  </doc>
  int iseed1, iseed2;
}

<doc> Designated initializer.  </doc>
id (self)
  initWithSeed (int, int) (s1, s2)
{
  iseed1 = s1;
  iseed2 = s2;
}

<doc> Default initializer.  </doc>
id
  init
{
  = [self initWithSeed (DEFSEED1, DEFSEED2)];
}

<doc> Short initializer.  </doc>
id
  initWithSeed int seed
{
  = [self initWithSeed (seed ^ DEFSEED1, seed ^ DEFSEED2)];
}

<doc> Initialize with the date {d}.  </doc>
id
  initWith Date d
{
  = [self initWithSeed int (1000.0 * [d timeIntervalSinceNow])];
}

double
  next
{
  // Does raw return something in the right range?
  // Tue May 23 21:12:18 2000, tiggr@home.tiggr.net
  = [self raw];
}

double
  raw
{
  int k, iz;

  k = iseed1 / 53688;
  iseed1 = 40014 * (iseed1 - k * 53668) - k * 12211;
  if (iseed1 < 0)
    iseed1 += 2147483563;

  k = iseed2 / 52774;
  iseed2 = 40692 * (iseed2 - k * 52774) - k * 3791;

  if (iseed2 < 0)
    iseed2 += 2147483399;

  iz = iseed1 - iseed2;

  if (iz < 1)
    iz += 2147483562;

  = double (iz) * 4.656613e-10;
}

long
  getSeed
{
  // Is this really the desired functionality?
  // Tue May 23 21:14:23 2000, tiggr@home.tiggr.net
  = (iseed1 * long (INT_MAX + iseed2));
}

end;

<doc> This is a Random number generator called Ranlux. A lot of the actual
    code was borrowed from the RngPack 1.0 Java package by Paul Houle.

    References:

    http://www.msc.cornell.edu/~houle/rngpack
    F. James, "Computer Physics Communications" 79 (1994) 111
    http://www.camk.edu.pl/~tomek/html.refs/ranlux.f90_2.html.  </doc>

implementation class
Ranlux: State, DoubleRandom
{
  <doc> Maximum luxury level.  </doc>
  const maxlev = 4;

  <doc> Default luxury level.  </doc>
  const lxdflt = 3;

  const igiga = 1000000000;
  const jsdflt = 314159265;

  const twop12 = 4096;
  const itwo24 = 1 << 24;
  const icons = 2147483563;
}

instance (id) (r)
  new
{
  r = [self alloc];

  int s = int (2.0e8 * [Date relativeTimeIntervalSinceNow]) ^ [r hashq];
  = [r initWithSeed s];
}

instance (id)
  newWithDefault
{
  = [[self alloc] init];
}

instance (id)
  newWithSeed int seed
{
  = [[self alloc] initWithSeed seed];
}

end;

implementation instance
Ranlux
{
  MutableIntArray iseeds;
  MutableIntArray isdext;

  MutableIntArray next;

  int luxlev;
  int nskip;
  int inseed;
  int jseed;

  int in24, kount, mkount, i24, j24;
  MutableFloatArray seeds;
  float carry;
  float twom24;
  float twom12;

  MutableIntArray ndskip;
}

id (self)
  init
{
  [self initLux];
  [self initArrays];
  [self rluxdef];
}

id (self)
  initWithSeed int ins
{
  [self initLux];
  [self initArrays];
  [self rluxgo (lxdflt, ins < 0 ? -ins : ins)];
}

id (self)
  initWithSeed int ins
	 atLux int lux
{
  [self initLux];
  [self initArrays];
  [self rluxgo (lux, ins < 0 ? -ins : ins)];
}

void
  initLux
{
  luxlev = lxdflt;
  in24 = 0;
  kount = 0;
  mkount = 0;
  i24 = 24;
  j24 = 10;

  carry = 0.0;

  ndskip = [MutableIntArray with (0, 24, 73, 199, 365)];
}

void
  initArrays
{
  /* Converted from Fortran: Fortran arrays start at 1, Java arrays start
     at 0 (TOM arrays do too).  Here we take the low road to
     compatibility...  We declare arrays that are one bigger than the
     Fortran code.  This wastes three ints and a double; If you're
     porting this to a Commodore 64 you might need the space.  */
  iseeds = [MutableIntArray withCapacity 24 + 1];
  isdext = [MutableIntArray withCapacity 25 + 1];
  next = [MutableIntArray withCapacity 24 + 1];
  seeds = [MutableFloatArray withCapacity 24 + 1];

  /* Initializing <array>[0]: */
  iseeds[0] = 0;
  isdext[0] = 0;
  next[0] = 0;
  seeds[0] = 0.0;
}

double
  next
{
  // Does raw return something in the right range?
  // Tue May 23 21:12:18 2000, tiggr@home.tiggr.net
  = [self raw];
}

double
  raw
{
  int i,k,lp;
  float uni, out;

  uni = seeds[j24] - seeds[i24] - carry;
  if (uni < 0.0)
    {
      uni += 1.0;
      carry = twom24;
    }
  else
    carry = 0.0;

  seeds[i24] = uni;

  i24 = next[i24];
  j24 = next[j24];

  out = uni;

  if (uni < twom12)
    out += twom24 * seeds[j24];

  /* Zero is forbidden in case user wants logarithms.  */
  if (out == 0.0)
    out = twom24 * twom24;

  in24++;

  if (in24 == 24)
    {
      in24 = 0;
      kount += nskip;
      for (i = 1; i <= nskip; i++)
	{
	  uni = seeds[j24] - seeds[i24] - carry;
	  if (uni < 0.0)
	    {
	      uni += 1.0;
	      carry = twom24;
            }
	  else
	    carry = 0.0;

	  seeds[i24] = uni;

	  i24 = next[i24];
	  j24 = next[j24];
        }
    }

  kount++;

  if (kount >= igiga)
    {
      mkount++;
      kount -= igiga;
    }

  return out;
}

void
  rluxdef
{
  int lp,i,k;

  jseed = jsdflt;
  inseed = jseed;
  luxlev = lxdflt;
  nskip = ndskip[luxlev];
  lp = nskip + 24;
  in24 = 0;
  kount = 0;
  mkount = 0;
  twom24 = 1.0;

  for (i = 1; i <= 24; i++)
    {
      twom24 = twom24 * 0.5;
      k = jseed / 53668;
      jseed = 40014 * (jseed - k * 53668) - k * 12211;
      if (jseed < 0)
	jseed = jseed + icons;

      iseeds[i] = jseed % itwo24;
    }

  twom12 = twom24 * 4096.0;

  for (i = 1 ; i <= 24; i++)
    {
      seeds[i] = iseeds[i] * twom24;
      next[i] = i - 1;
    }

  next[1] = 24;
  i24 = 24;
  j24 = 10;
  carry = 0.0;
  if (seeds[24] == 0.0)
    carry = twom24;
}

void
  rluxgo (int, int) (lux, ins)
{
  int ilx, i, iouter, isk, k, inner, izip, izip2;
  float uni;

  if (lux < 00)
    {
      luxlev = lxdflt;
    }
  else if (lux <= maxlev)
    {
      luxlev = lux;
    }
  else if (lux < 24 || lux > 2000)
    {
      luxlev = maxlev;
    }
  else
    {
      luxlev = lux;
      for (ilx = 0; ilx <= maxlev; ilx++)
	if (lux == ndskip[ilx] + 24)
	  luxlev = ilx;
    }

  if (luxlev <= maxlev)
    {
      nskip = ndskip[luxlev];
    }
  else
    {
      nskip = luxlev - 24;
    }

  in24 = 0;

  if (ins > 0)
    {
      jseed = ins;
    }
  else
    {
      jseed = jsdflt;
    }

  inseed = jseed;
  twom24 = 1.0;

  for (i = 1; i <= 24; i++)
    {
      twom24 = twom24 * 0.5;
      k = jseed / 53668;
      jseed = 40014 * (jseed - k * 53668) - k * 12211;

      if (jseed < 0)
	jseed = jseed + icons;
      iseeds[i] = jseed % itwo24;
    }

  twom12 = twom24 * 4096.0;

  for (i = 1; i <= 24; i++)
    {
      seeds[i] = iseeds[i] * twom24;
      next[i] = i - 1;
    }

  next[1] = 24;
  i24 = 24;
  j24 = 10;
  carry = 0.0;

  if (seeds[24] == 0.0)
    carry = twom24;

  kount = 0;
  mkount = 0;
}

end;

<doc> This is a Random number generator called Ranmar. A lot of the actual
    code was borrowed from the RngPack 1.0 Java package by Paul Houle.

    This is a quote from the "Ranmar.java" :

    "[Ranmar] is a lagged Fibonacci generator proposed by Marsaglia and
    Zaman and is a good research grade generator."

    References:

    http://www.msc.cornell.edu/~houle/rngpack.  </doc>

implementation class
Ranmar: DoubleRandom, State
{
  <doc> Default seed.  </doc>
  const DEFSEED=54217137;

  <doc> The 46,009,220th prime number, the largest prime less than
      9*10e8.  Used as a modulus because this version of
      {RANMAR} needs a seed between 0 and 9*10e8 and
      {BIG_PRIME} isn't commensurate with any regular period.
      </doc>
  const BIG_PRIME=899999963;
}

instance (id) (r)
   new
{
  r = [self alloc];

  int s = int (2.0e8 * [Date relativeTimeIntervalSinceNow]) ^ [r hashq];
  = [r initWithSeed s];
}

instance (id)
  newWithDefault
{
  = [[self alloc] init];
}

instance (id)
  newWithSeed int seed
{
  = [[self alloc] initWithSeed seed];
}

end;

implementation instance
Ranmar
{
  MutableDoubleArray u;
  MutableDoubleArray uvec;

  double c, cd, cm;

  int i97;
  int j97;
}

id (self)
  init
{
  [self ranmarin DEFSEED];
}

id (self)
  initWithSeed int seed
{
  seed = seed % BIG_PRIME;
  [self ranmarin seed < 0 ? -seed : seed];
}

id (self)
  initWith Date d
{
  int tim = int ([d timeIntervalSinceNow]);
  int tmp = tim % BIG_PRIME;
  int seed = tmp < 0 ? -tmp : tmp;
  [self ranmarin seed];
}

void
  ranmarin int ijkl
{
  int ij, kl, i, ii, j, jj, k, l, m;
  double s, t;

  u = [MutableDoubleArray withCapacity 97];
  uvec = [MutableDoubleArray withCapacity 97];

  ij = ijkl / 30082;
  kl = ijkl - 30082 * ij;

  i = ((ij / 177) % 177) + 2;
  j = (ij % 177) + 2;
  k = ((kl / 169) % 178) + 1;
  l = kl % 169;
  for (ii = 0; ii < 97; ii++)
    {
      s = 0.0;
      t = 0.5;
      for (jj = 0; jj < 24; jj++)
	{
	  m = (((i * j) % 179) * k) % 179;
	  i = j;
	  j = k;
	  k = m;
	  l = (53 * l + 1) % 169;
	  if (((l * m) % 64) >= 32)
	    s += t;
	  t *= 0.5;
	}
      u[ii] = s;
    }

  c   = 362436.0 / 16777216.0;
  cd  = 7654321.0 / 16777216.0;
  cm  = 16777213.0 / 16777216.0;
  i97 = 96;
  j97 = 32;
}

double
  next
{
  // Does raw return something in the right range?
  // Tue May 23 21:12:18 2000, tiggr@home.tiggr.net
  = [self raw];
}

double
  raw
{
  double uni;

  uni = u[i97] - u[j97];

  if (uni < 0.0)
    uni += 1.0;

  u[i97] = uni;

  if (--i97 < 0)
    i97 = 96;

  if (--j97 < 0)
    j97 = 96;

  c -= cd;

  if (c < 0.0)
    c += cm;

  uni -= c;

  if (uni < 0.0)
    uni += 1.0;

  = uni;
}

end;
